from typing import Mapping
from xml.dom import Node

def xml_to_dict(data: str | Node) -> Mapping[str, Mapping | str]: ...
def dict_to_xml(contents: Mapping[str, Mapping | str], encoding: str = ...) -> str: ...
